# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

"""
This package allows generating hashes for web assets. These hashes should be
attached to the URLs pointing to those assets. They can then be used to
reduce network traffic by sending smaller HTTP responses (i.e. 304 - not
modified) or sending the response directly from the configured cache folder.
"""
import fcntl
import hashlib
import io
import logging
import os
import re
import shutil
import textwrap
import time


log = logging.getLogger(__name__)


class VersionManager:
    """
    A handler for asset versioning via :term:`version strings
    <version string>`. The :func:`module initialization <score.webassets.init>`
    will instantiate an object of this class or a sub-class.

    The constructor argument *folder* must point to a valid path on the file
    system, which should **not** be regarded as a cache! Instances of this
    class will write files into this folder that will be needed during later
    HTTP requests.

    The only two relevant functions for the usage of this class are
    :meth:`.store` and :meth:`.load`.
    """

    hashregex = re.compile('^[0-9A-Fa-f]+$')

    def __init__(self, folder):
        self.folder = folder
        os.makedirs(folder, exist_ok=True)
        warning = os.path.join(folder, 'README.txt')
        open(warning, 'w').write(textwrap.dedent(
            'This folder is managed by a VersionManager '
            'of the python module score.webassets.\n\n'
            'Do not alter/delete any files!'
        ).strip())

    def store(self, category, path, hashers, content_generator):
        """
        Stores a version of an asset. The asset is specified as a 2-tuple of
        :term:`category <asset category>` and :term:`path <asset path>`,
        as usual.

        The hash functions, provided as parameter *hashers*, need to generate
        the exact same hashes as long as the content of this asset remains the
        same. If the *hashers* generate a previously generated hash, the
        function will return early. Otherwise the *content_generator* will be
        used to store the content of this version of the asset. That content
        will then be returned in consecutive calls to :meth:`.load`.

        See the :mod:`pyramid <score.webassets.pyramid>` package for an example
        use of this function.
        """
        if not hasattr(hashers, '__iter__'):
            hashers = hashers,
        hashes = []
        for hasher in hashers:
            hash = hasher()
            if hash:
                hashes.append(hash)
        if not len(hashes):
            return
        hash = hashlib.sha256(''.join(hashes).encode('UTF-8')).hexdigest()
        cachefile = self._cache_file(category, path, hash)
        if not os.path.isfile(cachefile):
            os.makedirs(os.path.dirname(cachefile), exist_ok=True)
            content = content_generator()
            open(cachefile, 'wb').write(content)
        return hash

    def load(self, category, path, hash):
        """
        Loads a previously stored hash of an asset described by
        :term:`category <asset category>` and :term:`path <asset path>`. This
        function will return a 2-tuple consisting of the content
        :term:`version <version string>` (denoted by the *hash*) and the age
        of the version.

        If the requested :term:`version <version string>` does not exist, this
        function will return `None` instead.

        See the :mod:`pyramid <score.webassets.pyramid>` package for an example
        use of this function.
        """
        cachefile = self._cache_file(category, path, hash)
        if not os.path.isfile(cachefile):
            return None
        content = open(cachefile, 'rb').read()
        return (content, time.time() - os.path.getmtime(cachefile))

    def create_file_hasher(self, files):
        """
        Returns a function that will generate a hash using the moste recent
        modification time in the *files* list.
        """
        if isinstance(files, str) or not hasattr(files, '__iter__'):
            files = files,
        if not len(files):
            return lambda: None
        return lambda: str(max(os.path.getmtime(f) for f in files))

    def _cache_file(self, category, path, hash):
        """
        Returns the file on the file system storing the content version of an
        asset.
        """
        assert VersionManager.hashregex.match(hash), "Invalid hash: %s" % hash
        return os.path.join(self.folder, category, path, hash)


class Dummy(VersionManager):
    """
    A Passthrough VersionManager that has no actual implementation. This class
    is the default VersionManager, if no other was configured.
    """

    def __init__(self, *args, **kwargs):
        pass

    def store(self, *args, **kwargs):
        pass

    def load(self, *args, **kwargs):
        pass


class Mercurial(VersionManager):
    """
    A version manager that ignores provided hashers and just uses the current
    changeset hashes of provided *repositories* folders.
    """

    def __init__(self, folder, repositories):
        super().__init__(folder)
        self.hashers = list(map(self.create_repo_hasher, repositories))

    def store(self, category, path, hashers, content_generator):
        # ignore *hashers*, use *self.hashers* instead
        return super().store(category, path, self.hashers, content_generator)

    def create_repo_hasher(self, repository):
        """
        Creates a hashing function for given mercurial *repository* folder.
        """
        def hasher():
            # the old implementation was using an external call to `hg`, which
            # was quite slow. now reading repository data directly, below.
            #   args = ['hg', '--debug', 'identify', '--id']
            #   process = subprocess.Popen(
            #       args,
            #       cwd=os.path.abspath(repository),
            #       stdin=subprocess.PIPE,
            #       stdout=subprocess.PIPE,
            #       stderr=subprocess.PIPE)
            #   stdout, stderr = process.communicate()
            #   if stderr:
            #       log.error('Error retrieving repository hash: %s' % stderr)
            #       return None
            #   return stdout[:-1]
            file = open(os.path.join(repository, '.hg', 'dirstate'), 'rb')
            bin = file.read(20)
            return hex(int.from_bytes(bin, 'big'))
        return hasher


class Netfs:
    """
    Wraps an existing VersionManager and uploads all files to :mod:`netfs
    <score.netfs>` to make asset versions available to peers.
    """

    # TODO: this is a bit of a mess. it should rather be implemented as some
    # kind of "backend" to a VersionManager.

    @staticmethod
    def _netfspath(category, path, hash):
        return 'webassets/%s/%s/%s' % (category, path, hash)

    def __init__(self, versionmanager, netfs):
        self.netfs = netfs
        self.versionmanager = versionmanager

    def __getattr__(self, attr):
        return getattr(self.versionmanager, attr)

    def store(self, category, path, hashers, content_generator):
        if not hasattr(hashers, '__iter__'):
            hashers = hashers,
        hashes = []
        for hasher in hashers:
            hash = hasher()
            if hash:
                hashes.append(hash)
        if not len(hashes):
            return
        hash = hashlib.sha256(''.join(hashes).encode('UTF-8')).hexdigest()
        cachepath = self._cache_file(category, path, hash)
        if os.path.isfile(cachepath):
            return hash
        os.makedirs(os.path.dirname(cachepath), exist_ok=True)
        tmppath = cachepath + '.tmp'
        tmpfile = open(tmppath, 'wb')
        connection = self.netfs.connect()
        from score.netfs import DownloadFailed
        try:
            fcntl.flock(tmpfile, fcntl.LOCK_EX)
            if os.path.exists(cachepath):
                # another process downloaded the file
                return hash
            time = connection.download(self._netfspath(category, path, hash),
                                       tmpfile)
        except DownloadFailed:
            # generate content and upload
            content = content_generator()
            open(cachepath, 'wb').write(content)
            netfspath = self._netfspath(category, path, hash)
            connection.upload(netfspath, io.BytesIO(content))
            # TODO: The next line commits netfs unconditionally,
            #   which is definitely not ok.
            connection.commit()
            return hash
        finally:
            fcntl.flock(tmpfile, fcntl.LOCK_UN)
            tmpfile.close()
        os.utime(tmppath, (time, time))
        shutil.move(tmppath, cachepath)
        return hash

    def load(self, category, path, hash):
        content = self.versionmanager.load(category, path, hash)
        if content is not None:
            return content
        cachepath = self._cache_file(category, path, hash)
        os.makedirs(os.path.dirname(cachepath), exist_ok=True)
        tmppath = cachepath + '.tmp'
        tmpfile = open(tmppath, 'wb')
        from score.netfs import DownloadFailed
        try:
            fcntl.flock(tmpfile, fcntl.LOCK_EX)
            if os.path.exists(cachepath):
                # another process downloaded the file
                return self.versionmanager.load(category, path, hash)
            time = self.netfs.connect().download(
                self._netfspath(category, path, hash), tmpfile)
        except DownloadFailed:
            return
        finally:
            fcntl.flock(tmpfile, fcntl.LOCK_UN)
            tmpfile.close()
        os.utime(tmppath, (time, time))
        shutil.move(tmppath, cachepath)
        return self.versionmanager.load(category, path, hash)



# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

"""
The aim of this module is to provide an infrastructure for handling assets -
like javascript or css files - in a web project. It is not intended to be
used directly; instead it provides footing for other modules. Those modules
making use of score.webassets' features should have a :term:`category
<asset category>` string, that can be used to distinguish different asset
types.

Currently it provides two features to be used by others - virtual assets and
asset versioning - as well as an adaption for the pyramid framework.
"""

import hashlib
from score.init import (
    init_object, init_cache_folder, ConfiguredModule, parse_bool)
from .versioning import Netfs as NetfsVersionManager


defaults = {
    'cachedir': None,
    'versionmanager': 'score.webassets.versioning.Dummy',
    'netfs': True,
}


def init(confdict, netfs_conf=None):
    """
    Initializes this module acoording to :ref:`our module initialization
    guidelines <module_initialization>` with the following configuration keys:

    :confkey:`cachedir` :faint:`[default=None]`
        A writable folder that will be used to cache intermediate values. This
        value is mostly unused in this module, but the initialized value can be
        used by other modules. The :mod:`css module <score.css>`, for example,
        will create a sub-folder beneath this folder, if it was initialized
        without an explicit `cachedir` of its own.

    :confkey:`versionmanager` :faint:`[default=score.webassets.versioning.Dummy]`
        The :class:`VersionManager` to use. This value will be converted to an
        object using :func:`score.init.init_object`.

        See the :mod:`package description <score.webassets.versioning>` for
        available implementations.

    :confkey:`netfs` :faint:`[default=True]`
        The initializer will upload all webassets to a :mod:`score.netfs`
        server, if one was configured. You can disable this feature by passing a
        `False` value here
    """
    conf = dict(defaults.items())
    conf.update(confdict)
    if conf['cachedir']:
        init_cache_folder(conf, 'cachedir', autopurge=True)
    versionmanager = init_object(conf, 'versionmanager')
    if netfs_conf and parse_bool(conf['netfs']):
        versionmanager = NetfsVersionManager(versionmanager, netfs_conf)
    return ConfiguredWebassetsModule(
        conf['cachedir'],
        versionmanager,
    )


class ConfiguredWebassetsModule(ConfiguredModule):
    """
    This module's :class:`configuration class
    <score.init.ConfiguredModule>`.
    """

    def __init__(self, cachedir, versionmanager):
        super().__init__(__package__)
        self.cachedir = cachedir
        self.versionmanager = versionmanager


class AssetNotFound(Exception):
    """
    Thrown when an asset was requested, but not found. Web applications might
    want to return the HTTP status code 404 in this case.

    Assets are uniquely identified by the combination of a :term:`category
    <asset category>` string and a :term:`path <asset path>`.
    """

    def __init__(self, category, path):
        self.category = category
        self.path = path

    def __str__(self):
        return '%s(%s)' % (self.path, self.category)


class VirtualAssets:
    """
    A Container for :term:`virtual assets <virtual asset>`, i.e. assets that
    have a :term:`path <asset path>`, but are not actually files. Virtual
    assets are instead registered with a generator function that generates
    its content. To be precise, each virtual asset consists of the following
    properties:

    - an :term:`asset path`,
    - a callback function that generates its content and
    - a hashing function that will generate the same string, as long as the
      content of the asset stays the same.

    The most convenient way to use this class is via an automatic
    :meth:`.decorator`.
    """

    def __init__(self):
        self.assets = {}

    def register(self, path, callback, hasher=None):
        """
        Registers a new :term:`virtual asset` beneath the provided *path*. The
        *callback* will be used to generate the content of the asset.

        The optional *hasher* callback must return the same string as long as
        the *callback* generates the same content. If no *hasher* is provided,
        one will be generated automatically that hashes the generated content.
        """
        assert path not in self.assets
        if hasher is None:
            def hasher():
                result = callback()
                if isinstance(result, str):
                    result = result.encode('UTF-8')
                sha = hashlib.sha256()
                sha.update(result)
                return sha.hexdigest()
        self.assets[path] = (callback, hasher)

    def paths(self):
        """
        Returns an iterator of all registered asset paths.
        """
        return self.assets.keys()

    def render(self, path):
        """
        Invokes the callback of the asset with given *path* and returns its
        result.
        """
        return self.assets[path][0]()

    def hash(self, path):
        """
        Invokes the hashing function of the asset with given *path* and
        returns its result.
        """
        return self.assets[path][1]()

    def decorator(self, suffix):
        """
        Creates a decorator that can be used to register callbacks
        conveniently:

        >>> v = VirtualAssets()
        >>> virtualcss = v.decorator('css')
        >>> @virtualcss
        ... def textcolor():
        ...   return 'body {color: black}'
        ...
        >>> @virtualcss('background-color.css')
        ... def bgcolor():
        ...   return 'body {background-color: white}'
        ...
        >>> def hasher():
        ...   try:
        ...     return os.path.getmtime('parrot.png')
        ...   except FileNotFoundError:
        ...     return 'dead'
        ...
        >>> @virtualcss('parrot.css', hasher)
        ... def parrot():
        ...   url = 'parrot.png?_=%s' % hasher()
        ...   return '.parrot {background-image: url(%s)}' % url
        ...
        >>> list(v.paths())
        ['textcolor.css', 'background-color.css', 'parrot.css']
        """
        class Decorator:
            def __new__(cls, *args):
                if callable(args[0]):
                    name = '%s.%s' % (args[0].__name__, suffix)
                    self.register(name, args[0])
                    return args[0]
                o = super().__new__(cls)
                o.name = args[0]
                o.hasher = None
                if len(args) > 1:
                    o.hasher = args[1]
                return o
            def __call__(o, callback):
                self.register(o.name, callback, o.hasher)
        return Decorator



# Copyright © 2015 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

"""
This package :ref:`integrates <framework_integration>` the module with
pyramid.

It registers a handler for the exception :exc:`AssetNotFound
<score.webassets.AssetNotFound>`, which returns the HTTP status code
``404 - Not found``.
"""

import os
from pyramid.request import Request
import score.webassets
import time


def assetnotfound(exc, request):
    """
    Returns an HTTP response with status code 404. This method is registered
    in the pyramid-specific :func:`init` function.
    """
    request.response.status = 404
    return request.response


def init(confdict, configurator, netfs_conf=None):
    """
    Performs the following steps:

    - Wraps the versionmanager in a PyramidVersionManager object to provide
      additional features.
    - Generates a dummy pyramid request (see below) and adds it to the
      configuration object as ``dummy_request``.
    - Registers the view assetnotfound for a handler to the
      :exc:`AssetNotFound` Exception.

    In addition to the configuration keys of the generic :func:`initializer
    function <score.webassets.init>`, this function additionally handles the
    following configuration keys:

    :confkey:`dummy_request` :faint:`[optional]`
        A pyramid request that can will be used to create URLs to assets. If
        this value is missing, a new :class:`DummyRequest` will be created
        automatically.
    """
    webconf = score.webassets.init(confdict, netfs_conf)
    webconf.versionmanager = PyramidVersionManager(webconf.versionmanager)
    if 'dummy_request' in confdict:
        assert isinstance(confdict['dummy_request'], Request)
        webconf.dummy_request = confdict['dummy_request']
    else:
        webconf.dummy_request = DummyRequest.blank('/', base_url='')
    webconf.dummy_request.registry = configurator.registry
    configurator.add_view(assetnotfound, context=score.webassets.AssetNotFound)
    return webconf


class DummyRequest(Request):
    """
    Derives from :class:`pyramid.request.Request` to override url generating
    functions. By default, this class will omit the schema/domain of the url,
    making all URLs relative to the current host. This behaviour can be changed
    by setting the attribute ``app_url`` on the request object:

    >>> d = DummyRequest.blank('/')
    >>> d.route_url('somewhere')
    /somewhere
    >>> d.app_url = 'http://foo.bar'
    >>> d.route_url('somewhere')
    http://foo.bar/somewhere
    """

    def __init__(self, *args, **kwargs):
        self.app_url = ''
        Request.__init__(self, *args, **kwargs)

    def route_url(self, route_name, *elements, **kw):
        if '_app_url' not in kw:
            kw['_app_url'] = self.app_url
        return super().route_url(route_name, *elements, **kw)


class PyramidVersionManager:
    """
    A wrapper for other :class:`VersionManagers <VersionManager>` that adds a
    function for handling version hashes in the request url.
    """

    def __init__(self, versionmanager):
        self.versionmanager = versionmanager

    def __getattr__(self, attr):
        return getattr(self.versionmanager, attr)

    def handle_pyramid_request(self, category, path, request):
        """
        Tries to populate the :attr:`response
        <pyramid.request.Request.response>` of the pyramid request with the
        cached version of an asset as described in the introduction to
        :mod:`score.webassets.versioning`. The asset is specified as a
        combination of :term:`category <asset category>` and :term:`path
        <asset path>`, as usual.

        This function will handle the headers `If-None-Match` and
        `If-Modified-Since` with a status code of 304 if the according asset
        version exists and respond with the cached content if the request
        provides a :term:`version string` via the GET value ``_v``.

        If this function manages to populate the response object of the
        *request* with the appropriate values for sending to the client, it
        will return `True`. If this function cannot find the correct asset
        version, or the client did not send any version information, it will
        not populate the response and return `False`.
        """
        if 'If-None-Match' in request.headers:
            hash = request.headers['If-None-Match'].strip('"')
            cachefile = self._cache_file(category, path, hash)
            if os.path.isfile(cachefile):
                request.response.status = 304
                return request.response
        if 'If-Modified-Since' in request.headers:
            import email.utils as eut
            t = time.mktime(eut.parsedate(request.headers['If-Modified-Since']))
            folder = os.path.join(self.cachedir, category, path)
            for f in os.listdir(folder):
                if os.path.getmtime(f) > t:
                    modified = True
                    break
            if not modified:
                request.response.status = 304
                return True
        if '_v' in request.GET:
            hash = request.GET['_v']
            result = self.versionmanager.load(category, path, hash)
            if result:
                body, age = result
                request.response.body = body
                request.response.max_age = str(60 * 60 * 24 * 30 * 12)  # 1 year
                request.response.age = age
                request.response.etag = hash
                return True
        return False
